#include "abcdisk.h"

const char *program;

void *xmalloc(size_t len)
{
    void *p = malloc(len);

    if (!p) {
        perror(program);
        exit(1);
    }

    return p;
}

void *xzalloc(size_t len)
{
    void *p = calloc(len, 1);

    if (!p) {
        perror(program);
        exit(1);
    }

    return p;
}

struct disk *allocate_disk(const struct diskfmt *fmt)
{
    struct disk *disk = xzalloc(sizeof(struct disk));
    size_t size = fmt->sectors << 8;
    int i, starter;
    int secperclust = 1 << fmt->cluster_shift;
    unsigned char *p;

    disk->fmt = fmt;
    disk->image = xmalloc(size);
    disk->bitmap = disk->image + (fmt->syssec << 8);
    disk->mfd.disk = disk;
    disk->mfd.bitmap = disk->bitmap;
    disk->mfd.dirsec = disk->image + (fmt->dirsec[0] << 8);
    disk->mfd.block = 0;        /* This is the MFD */

    /* Post-formatting content? */
    memset(disk->image, 0x40, size);

    /* Create empty bitmap */
    memset(disk->bitmap, 0, 256);
    for (i = fmt->sectors >> fmt->cluster_shift; i < 239 * 8; i++)
        mark_used(disk, i);     /* Mark nonexistent clusters busy */

    /* 239..254 is the MFD bitmap */
    memset(&disk->bitmap[239], 0x01, fmt->dirsecs);

    /* This is claimed in a message found on ABC-klubben MSG */
    disk->bitmap[0xff] = fmt->cluster_shift;

    /* Mark all the system sectors used */
    starter = ((fmt->dirsec[0] > fmt->dirsec[1])
	       ? fmt->dirsec[0] : fmt->dirsec[1]) + fmt->dirsecs;
    starter = (starter + secperclust - 1) >> fmt->cluster_shift;
    for (i = 0; i < starter; i++)
        mark_used(disk, i);

    /* Copy bitmap to backup empty bitmap (= bad block map) */
    memcpy(disk->bitmap + 256, disk->bitmap, 256);

    /* Create MFD directory sectors */
    p = disk->mfd.dirsec;
    for (i = 0; i < fmt->dirsecs; i++) {
      memset(p, 0x00, 16);	/* Reserved entry */
      memset(p + 16, 0xff, 256-16);
      p += 256;
    }

    return disk;
}

void free_disk(struct disk *disk)
{
    if (disk) {
	free(disk->image);
	free(disk);
    }
}
